@@ -3,7 +3,7 @@
  * pgpool: a language independent connection pool server for PostgreSQL
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2022	PgPool Global Development Group
+ * Copyright (c) 2003-2020	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -187,9 +187,18 @@ connection_do_auth(POOL_CONNECTION_POOL_SLOT * cp, char *password)
 
 		buf = palloc0(2 * (MD5_PASSWD_LEN + 4));	/* hash + "md5" + '\0' */
 
-		/* build md5 password */
+		/* set buffer address for building md5 password */
 		buf1 = buf + MD5_PASSWD_LEN + 4;
-		pool_md5_encrypt(password, cp->sp->user, strlen(cp->sp->user), buf1);
+
+		/*
+		 * If the supplied password is already in md5 hash format, we just
+		 * copy it. Otherwise calculate the md5 hash value.
+		 */
+		if (!strncmp("md5", password, 3) && (strlen(password) - 3) == MD5_PASSWD_LEN)
+			memcpy(buf1, password + 3, MD5_PASSWD_LEN + 1);
+		else
+			pool_md5_encrypt(password, cp->sp->user, strlen(cp->sp->user), buf1);
+
 		pool_md5_encrypt(buf1, salt, 4, buf + 3);
 		memcpy(buf, "md5", 3);
 
@@ -2030,12 +2039,9 @@ pool_read_message_length(POOL_CONNECTION_POOL * cp)
 				 errdetail("slot: %d length: %d", i, length)));
 
 		if (length != length0)
-		{
-			ereport(LOG,
+			ereport(ERROR,
 					(errmsg("unable to read message length"),
 					 errdetail("message length (%d) in slot %d does not match with slot 0(%d)", length, i, length0)));
-			return -1;
-		}
 
 	}
 
@@ -2060,14 +2066,14 @@ pool_read_message_length2(POOL_CONNECTION_POOL * cp)
 	int			i;
 	static int	length_array[MAX_CONNECTION_SLOTS];
 
-	/* read message from main node */
+	/* read message from master node */
 	pool_read(CONNECTION(cp, MASTER_NODE_ID), &length0, sizeof(length0));
 
 	length0 = ntohl(length0);
 	length_array[MASTER_NODE_ID] = length0;
 	ereport(DEBUG5,
 			(errmsg("reading message length"),
-			 errdetail("main slot: %d length: %d", MASTER_NODE_ID, length0)));
+			 errdetail("master slot: %d length: %d", MASTER_NODE_ID, length0)));
 
 	for (i = 0; i < NUM_BACKENDS; i++)
 	{
@@ -2078,11 +2084,11 @@ pool_read_message_length2(POOL_CONNECTION_POOL * cp)
 			length = ntohl(length);
 			ereport(DEBUG5,
 					(errmsg("reading message length"),
-					 errdetail("main slot: %d length: %d", i, length)));
+					 errdetail("master slot: %d length: %d", i, length)));
 
 			if (length != length0)
 			{
-				ereport(DEBUG1,
+				ereport(LOG,
 						(errmsg("reading message length"),
 						 errdetail("message length (%d) in slot %d does not match with slot 0(%d)", length, i, length0)));
 			}
@@ -2101,41 +2107,6 @@ pool_read_message_length2(POOL_CONNECTION_POOL * cp)
 	return &length_array[0];
 }
 
-/*
- * By given message length array, emit log message to complain the difference.
- * If no difference, no log is emitted.
- * If "name" is not NULL, it is added to the log message.
- */
-void
-pool_emit_log_for_message_length_diff(int *length_array, char *name)
-{
-	int			length0,	/* message length of main node id */
-				length;
-	int			i;
-
-	length0 = length_array[MASTER_NODE_ID];
-
-	for (i = 0; i < NUM_BACKENDS; i++)
-	{
-		if (VALID_BACKEND(i))
-		{
-			length = length_array[i];
-
-			if (length != length0)
-			{
-				if (name != NULL)
-					ereport(LOG,
-							(errmsg("ParameterStatus \"%s\": node %d message length %d is different from main node message length %d",
-									name, i, length_array[i], length0)));
-				else
-					ereport(LOG,
-							(errmsg("node %d message length %d is different from main node message length %d",
-									i, length_array[i], length0)));
-			}
-		}
-	}
-}
-
 signed char
 pool_read_kind(POOL_CONNECTION_POOL * cp)
 {
