@@ -5,7 +5,7 @@
  * pgpool: a language independent connection pool server for PostgreSQL
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2023	PgPool Global Development Group
+ * Copyright (c) 2003-2019	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -68,7 +68,7 @@ static RETSIGTYPE reload_config_handler(int sig);
 static RETSIGTYPE authentication_timeout(int sig);
 static void send_params(POOL_CONNECTION * frontend, POOL_CONNECTION_POOL * backend);
 static void send_frontend_exits(void);
-static void connection_count_up(void);
+static int	connection_count_up(void);
 static void connection_count_down(void);
 static bool connect_using_existing_connection(POOL_CONNECTION * frontend,
 								  POOL_CONNECTION_POOL * backend,
@@ -137,13 +137,7 @@ do_child(int *fds)
 	struct timeval timeout;
 	static int	connected = 0;	/* non 0 if has been accepted connections from
 								 * frontend */
-
-	/* counter for child_max_connections.  "volatile" declaration is necessary
-	 * so that this is counted up even if long jump is issued due to
-	 * ereport(ERROR).
-	 */
-	volatile	int			connections_count = 0;
-
+	int			connections_count = 0;	/* used if child_max_connections > 0 */
 	char		psbuf[NI_MAXHOST + 128];
 
 	ereport(DEBUG2,
@@ -300,6 +294,7 @@ do_child(int *fds)
 		StartupPacket *sp;
 		int			front_end_fd;
 		SockAddr	saddr;
+		int			con_count;
 
 		/* reset per iteration memory context */
 		MemoryContextSwitchTo(ProcessLoopContext);
@@ -330,7 +325,32 @@ do_child(int *fds)
 		if (front_end_fd == RETRY)
 			continue;
 
-		connection_count_up();
+		/*
+		 * Check if max connections from clients execeeded.
+		 */
+		con_count = connection_count_up();
+		if (con_count > (pool_config->num_init_children - pool_config->reserved_connections))
+		{
+			POOL_CONNECTION * cp;
+			cp = pool_open(front_end_fd, false);
+			if (cp == NULL)
+			{
+				connection_count_down();
+				continue;
+			}
+			connection_count_down();
+			pool_send_fatal_message(cp, 3, "53300",
+									"Sorry, too many clients already",
+									"",
+									"",
+									__FILE__, __LINE__);
+			ereport(ERROR,
+					(errcode(ERRCODE_TOO_MANY_CONNECTIONS),
+					 errmsg("Sorry, too many clients already")));
+			pool_close(cp);
+			continue;
+		}
+
 		accepted = 1;
 
 		check_config_reload();
@@ -485,10 +505,8 @@ backend_cleanup(POOL_CONNECTION * volatile *frontend, POOL_CONNECTION_POOL * vol
 	if (cache_connection)
 	{
 		/*
-		 * For those special databases, or when frontend client exits
-		 * abnormally, we don't cache connection to backend.  Also we have to
-		 * check whether reset query failed. If so, the existing connection to
-		 * backend may not be used and we don't want to cache the connection.
+		 * For those special databases, and when frontend client exits
+		 * abnormally, we don't cache connection to backend.
 		 */
 		if ((sp &&
 			 (!strcmp(sp->database, "template0") ||
@@ -497,8 +515,7 @@ backend_cleanup(POOL_CONNECTION * volatile *frontend, POOL_CONNECTION_POOL * vol
 			  !strcmp(sp->database, "regression"))) ||
 			(*frontend != NULL &&
 			 ((*frontend)->socket_state == POOL_SOCKET_EOF ||
-			  (*frontend)->socket_state == POOL_SOCKET_ERROR)) ||
-			reset_query_error)
+			  (*frontend)->socket_state == POOL_SOCKET_ERROR)))
 			cache_connection = false;
 	}
 
@@ -506,12 +523,8 @@ backend_cleanup(POOL_CONNECTION * volatile *frontend, POOL_CONNECTION_POOL * vol
 	 * Close frontend connection
 	 */
 	reset_connection();
-
-	if (*frontend != NULL)
-	{
-		pool_close(*frontend);
-		*frontend = NULL;
-	}
+	pool_close(*frontend);
+	*frontend = NULL;
 
 	if (cache_connection == false)
 	{
@@ -660,11 +673,7 @@ read_startup_packet(POOL_CONNECTION * cp)
 				}
 				else
 				{
-					ereport(DEBUG1,
-							(errmsg("reading startup packet"),
-							 errdetail("guc name: %s value: %s", p, p+strlen(p)+1)));
 					p += (strlen(p) + 1);
-
 				}
 
 				p += (strlen(p) + 1);
@@ -1049,10 +1058,10 @@ static RETSIGTYPE die(int sig)
 {
 	int			save_errno;
 
-	save_errno = errno;
-
 	POOL_SETMASK(&BlockSig);
 
+	save_errno = errno;
+
 #ifdef NOT_USED
 	ereport(LOG,
 			(errmsg("child process received shutdown request signal %d", sig)));
@@ -1097,9 +1106,8 @@ static RETSIGTYPE die(int sig)
 			break;
 	}
 
-	POOL_SETMASK(&UnBlockSig);
-
 	errno = save_errno;
+	POOL_SETMASK(&UnBlockSig);
 }
 
 /*
@@ -1294,7 +1302,8 @@ child_will_go_down(int code, Datum arg)
 	if (accepted)
 		connection_count_down();
 
-	if (pool_config->memory_cache_enabled && !pool_is_shmem_cache())
+	if ((pool_config->memory_cache_enabled || pool_config->enable_shared_relcache)
+		&& !pool_is_shmem_cache())
 	{
 		memcached_disconnect();
 	}
@@ -1534,10 +1543,11 @@ discard_persistent_db_connection(POOL_CONNECTION_POOL_SLOT * cp)
 }
 
 /*
- * Count up connection counter (from frontend to pgpool)
- * in shared memory
+ * Count up connection counter (from frontend to pgpool) in shared memory and
+ * returns current counter value.  Please note that the returned value may not
+ * be up to date since locking has been already released.
  */
-static void
+static int
 connection_count_up(void)
 {
 	pool_sigset_t oldmask;
@@ -1545,8 +1555,10 @@ connection_count_up(void)
 	POOL_SETMASK2(&BlockSig, &oldmask);
 	pool_semaphore_lock(CONN_COUNTER_SEM);
 	Req_info->conn_counter++;
+	elog(DEBUG5, "connection_count_up: number of connected children: %d", Req_info->conn_counter);
 	pool_semaphore_unlock(CONN_COUNTER_SEM);
 	POOL_SETMASK(&oldmask);
+	return Req_info->conn_counter;
 }
 
 /*
@@ -1570,6 +1582,7 @@ connection_count_down(void)
 	 */
 	if (Req_info->conn_counter > 0)
 		Req_info->conn_counter--;
+	elog(DEBUG5, "connection_count_down: number of connected children: %d", Req_info->conn_counter);
 	pool_semaphore_unlock(CONN_COUNTER_SEM);
 	POOL_SETMASK(&oldmask);
 }
@@ -1636,7 +1649,7 @@ select_load_balancing_node(void)
 							   pool_config->db_redirect_tokens->token[index_db].weight_token)));
 
 			tmp = choose_db_node_id(pool_config->db_redirect_tokens->token[index_db].right_token);
-			if (tmp == -1 || (tmp >= 0 && VALID_BACKEND_RAW(tmp)))
+			if (tmp == -1 || (tmp >= 0 && VALID_BACKEND(tmp)))
 				suggested_node_id = tmp;
 		}
 	}
@@ -1677,7 +1690,7 @@ select_load_balancing_node(void)
 								   pool_config->app_name_redirect_tokens->token[index_app].weight_token)));
 
 				tmp = choose_db_node_id(pool_config->app_name_redirect_tokens->token[index_app].right_token);
-				if (tmp == -1 || (tmp >= 0 && VALID_BACKEND_RAW(tmp)))
+				if (tmp == -1 || (tmp >= 0 && VALID_BACKEND(tmp)))
 					suggested_node_id = tmp;
 			}
 		}
@@ -1722,7 +1735,7 @@ select_load_balancing_node(void)
 
 	for (i = 0; i < NUM_BACKENDS; i++)
 	{
-		if (VALID_BACKEND(i))
+		if (VALID_BACKEND_RAW(i))
 		{
 			if (i == no_load_balance_node_id)
 				continue;
@@ -1748,7 +1761,7 @@ select_load_balancing_node(void)
 		if ((suggested_node_id == -1 && i == PRIMARY_NODE_ID) || i == no_load_balance_node_id)
 			continue;
 
-		if (VALID_BACKEND(i) && BACKEND_INFO(i).backend_weight > 0.0)
+		if (VALID_BACKEND_RAW(i) && BACKEND_INFO(i).backend_weight > 0.0)
 		{
 			if (r >= total_weight)
 				selected_slot = i;
@@ -1822,7 +1835,7 @@ check_restart_request(void)
 	if (pool_get_my_process_info()->need_to_restart)
 	{
 		ereport(LOG,
-				(errmsg("failback event detected"),
+				(errmsg("failover or failback event detected"),
 				 errdetail("restarting myself")));
 
 		pool_get_my_process_info()->need_to_restart = 0;
@@ -1894,50 +1907,7 @@ wait_for_new_connections(int *fds, struct timeval *timeout, SockAddr *saddr)
 	 */
 	if (SERIALIZE_ACCEPT)
 	{
-		if (pool_config->connection_life_time == 0)
-		{
-			pool_semaphore_lock(ACCEPT_FD_SEM);
-		}
-		else
-		{
-			int sts;
-
-			for (;;)
-			{
-				sts = pool_semaphore_lock_allow_interrupt(ACCEPT_FD_SEM);
-
-				/* Interrupted by alarm */
-				if (sts == -2)
-				{
-					/*
-					 * Check if there are expired connection_life_time.
-					 */
-					if (backend_timer_expired)
-					{
-						/*
-						 * We add 10 seconds to connection_life_time so that there's
-						 * enough margin.
-						 */
-						int	seconds = pool_config->connection_life_time + 10;
-
-						while (seconds-- > 0)
-						{
-							/* check backend timer is expired */
-							if (backend_timer_expired)
-							{
-								pool_backend_timer();
-								backend_timer_expired = 0;
-								break;
-							}
-							sleep(1);
-						}
-					}
-				}
-				else	/* success or other error */
-					break;
-			}
-		}
-
+		pool_semaphore_lock(ACCEPT_FD_SEM);
 		set_ps_display("wait for connection request", false);
 		ereport(DEBUG1,
 				(errmsg("LOCKING select()")));
@@ -2003,7 +1973,7 @@ wait_for_new_connections(int *fds, struct timeval *timeout, SockAddr *saddr)
 			return RETRY;
 		ereport(ERROR,
 				(errmsg("failed to accept user connection"),
-				 errdetail("select on socket failed with error : \"%m\"")));
+				 errdetail("select on socket failed with error : \"%s\"", strerror(errno))));
 	}
 
 	/* timeout */
@@ -2062,7 +2032,7 @@ retry_accept:
 		if (errno != EAGAIN && errno != EWOULDBLOCK && errno != EINTR)
 			ereport(ERROR,
 					(errmsg("failed to accept user connection"),
-					 errdetail("accept on socket failed with error : \"%m\"")));
+					 errdetail("accept on socket failed with error : \"%s\"", strerror(errno))));
 		return RETRY;
 
 	}
@@ -2079,7 +2049,7 @@ retry_accept:
 					   sizeof(on)) < 0)
 		{
 			ereport(WARNING,
-					(errmsg("wait_for_new_connections: setsockopt failed with error \"%m\"")));
+					(errmsg("wait_for_new_connections: setsockopt failed with error \"%s\"", strerror(errno))));
 			close(afd);
 			return -1;
 		}
@@ -2168,7 +2138,7 @@ validate_backend_connectivity(int front_end_fd)
 		if (front_end_fd > 0)
 		{
 			POOL_CONNECTION *cp;
-			StartupPacket *volatile sp = NULL;
+			StartupPacket *volatile sp;
 
 			/*
 			 * we do not want to report socket error, as above errors will be
@@ -2192,15 +2162,17 @@ validate_backend_connectivity(int front_end_fd)
 										error_hint,
 										__FILE__,
 										__LINE__);
+				
 			}
 			PG_CATCH();
 			{
+				pool_free_startup_packet(sp);
 				FlushErrorState();
 				ereport(FATAL,
 						(errmsg("%s", error_msg), errdetail("%s", error_detail), errhint("%s", error_hint)));
-				pfree(sp);
 			}
 			PG_END_TRY();
+			pool_free_startup_packet(sp);
 		}
 		ereport(FATAL,
 				(errmsg("%s", error_msg), errdetail("%s", error_detail), errhint("%s", error_hint)));
@@ -2246,12 +2218,12 @@ get_connection(int front_end_fd, SockAddr *saddr)
 					   sizeof(on)) < 0)
 			ereport(ERROR,
 					(errmsg("failed to accept user connection"),
-					 errdetail("setsockopt on socket failed with error : \"%m\"")));
+					 errdetail("setsockopt on socket failed with error : \"%s\"", strerror(errno))));
 
 		if (setsockopt(front_end_fd, SOL_SOCKET, SO_KEEPALIVE, (char *) &on, sizeof(on)) < 0)
 			ereport(FATAL,
 					(errmsg("failed to accept user connection"),
-					 errdetail("setsockopt on socket failed with error : \"%m\"")));
+					 errdetail("setsockopt on socket failed with error : \"%s\"", strerror(errno))));
 
 	}
 
@@ -2260,7 +2232,7 @@ get_connection(int front_end_fd, SockAddr *saddr)
 		close(front_end_fd);
 		ereport(ERROR,
 				(errmsg("failed to accept user connection"),
-				 errdetail("unable to open connection with remote end : \"%m\"")));
+				 errdetail("unable to open connection with remote end : \"%s\"", strerror(errno))));
 	}
 
 	/* save ip address for hba */
@@ -2287,7 +2259,6 @@ retry_startup:
 	{
 		cancel_request((CancelPacket *) sp->startup_packet);
 		pool_free_startup_packet(sp);
-		connection_count_down();
 		return NULL;
 	}
 
@@ -2303,20 +2274,6 @@ retry_startup:
 		goto retry_startup;
 	}
 
-	/* GSSAPI? */
-	if (sp->major == 1234 && sp->minor == 5680)
-	{
-		ereport(DEBUG1,
-				(errmsg("selecting backend connection"),
-				 errdetail("GSSAPI request from client")));
-
-		/* sorry, Pgpool-II does not support GSSAPI yet */
-		pool_write_and_flush(frontend, "N", 1);
-
-		pool_free_startup_packet(sp);
-		goto retry_startup;
-	}
-
 	frontend->protoVersion = sp->major;
 	frontend->database = pstrdup(sp->database);
 	frontend->username = pstrdup(sp->user);
@@ -2353,7 +2310,7 @@ retry_startup:
 	{
 		ereport(LOG,
 				(errmsg("selecting backend connection"),
-				 errdetail("failback event detected, discarding existing connections")));
+				 errdetail("failover or failback event detected, discarding existing connections")));
 
 		pool_get_my_process_info()->need_to_restart = 0;
 		close_idle_connection(0);
