@@ -341,8 +341,12 @@ pool_connection_pool_timer(POOL_CONNECTION_POOL * backend)
 			(errmsg("setting backend connection close timer"),
 			 errdetail("close time %ld", time(NULL))));
 
-	MASTER_CONNECTION(backend)->closetime = time(NULL); /* set connection close
-														 * time */
+	/* Set connection close time */
+	for (i = 0; i < NUM_BACKENDS; i++)
+	{
+		if (CONNECTION_SLOT(backend, i))
+			CONNECTION_SLOT(backend, i)->closetime = time(NULL);
+	}
 
 	if (pool_config->connection_life_time == 0)
 		return;
@@ -851,6 +855,7 @@ static POOL_CONNECTION_POOL * new_connection(POOL_CONNECTION_POOL * p)
 	POOL_CONNECTION_POOL_SLOT *s;
 	int			active_backend_count = 0;
 	int			i;
+	bool		status_changed = false;
 
 	MemoryContext oldContext = MemoryContextSwitchTo(TopMemoryContext);
 
@@ -892,6 +897,8 @@ static POOL_CONNECTION_POOL * new_connection(POOL_CONNECTION_POOL * p)
 
 		if (create_cp(s, i) == NULL)
 		{
+			pfree(s);
+
 			/*
 			 * If failover_on_backend_error is true, do failover. Otherwise,
 			 * just exit this session or skip next health node.
@@ -946,21 +953,25 @@ static POOL_CONNECTION_POOL * new_connection(POOL_CONNECTION_POOL * p)
 			}
 			child_exit(POOL_EXIT_AND_RESTART);
 		}
-
-		p->info[i].create_time = time(NULL);
-		p->slots[i] = s;
-
-		pool_init_params(&s->con->params);
-
-		if (BACKEND_INFO(i).backend_status != CON_UP)
+		else
 		{
-			BACKEND_INFO(i).backend_status = CON_UP;
-			pool_set_backend_status_changed_time(i);
+			p->info[i].create_time = time(NULL);
+			p->slots[i] = s;
+
+			pool_init_params(&s->con->params);
+
+			if (BACKEND_INFO(i).backend_status != CON_UP)
+			{
+				BACKEND_INFO(i).backend_status = CON_UP;
+				pool_set_backend_status_changed_time(i);
+				status_changed = true;
+			}
+			active_backend_count++;
 		}
-		active_backend_count++;
 	}
 
-	(void) write_status_file();
+	if (status_changed)
+		(void) write_status_file();
 
 	MemoryContextSwitchTo(oldContext);
 
