@@ -23,16 +23,17 @@ do
 	export PGPORT=$PGPOOL_PORT
 
 
+	echo "log_min_messages = 'debug5'" >> etc/pgpool.conf
 	# start pgpool-II
 	./startall
 
-	wait_for_pgpool_startup
+	sleep 1
 
-	$PSQL test -p $PGPORT -c "SELECT pg_sleep(20);" &
+	$PSQL test -p $PGPORT -c "SELECT pg_sleep(10);" &
 
 	sleep 2
 	# get process id which query is executed
-	PID=`ps -efw |grep "postgres:" |grep SELECT | awk 'NR==1 {print $2}'`
+	PID=`ps -ef |grep "postgres:" |grep SELECT | awk 'NR==1 {print $2}'`
 
 	$PSQL test -p $PGPORT -c "SELECT pg_terminate_backend($PID)"
 
@@ -52,6 +53,19 @@ do
 		exit 1
 	fi
 
+	sleep 5
+
+	COUNT_UP=`grep "connection_count_up" log/pgpool.log |wc -l`
+	COUNT_DOWN=`grep "connection_count_down" log/pgpool.log |wc -l`
+
+	echo "count_up: ${COUNT_UP}"
+	echo "count_down: ${COUNT_DOWN}"
+	if [ ${COUNT_UP} != ${COUNT_DOWN} ];then
+		echo "connection count is not matched."
+		./shutdownall
+		exit 1
+	fi
+
 	./shutdownall
 
 	cd ..
